###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
import argparse

from rich import print

from nucs.constants import LOG_LEVEL_INFO, LOG_LEVELS
from nucs.examples.schur_lemma.schur_lemma_problem import SchurLemmaProblem
from nucs.solvers.backtrack_solver import BacktrackSolver

# Run with the following command (the second run is much faster because the code has been compiled):
# NUMBA_CACHE_DIR=.numba/cache python -m nucs.examples.schur_lemma -n 20 --symmetry_breaking
if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--log_level", choices=LOG_LEVELS, default=LOG_LEVEL_INFO)
    parser.add_argument("-n", type=int, default=20)
    parser.add_argument("--symmetry_breaking", action=argparse.BooleanOptionalAction, default=True)
    args = parser.parse_args()
    problem = SchurLemmaProblem(args.n, args.symmetry_breaking)
    solver = BacktrackSolver(problem, log_level=args.log_level, stacks_max_height=128)
    solver.solve_all()
    print(solver.get_statistics())



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
import numpy as np
from numba import njit  # type: ignore
from numpy.typing import NDArray

from nucs.constants import (
    EVENT_MASK_MAX,
    EVENT_MASK_MIN,
    MAX,
    MIN,
    PROP_CONSISTENCY,
    PROP_ENTAILMENT,
    PROP_INCONSISTENCY,
)


def get_complexity_min_geq(n: int, parameters: NDArray) -> float:
    """
    Returns the time complexity of the propagator as a float.
    :param n: the number of variables
    :param parameters: the parameters, unused here
    :return: a float
    """
    return 3 * n


def get_triggers_min_geq(n: int, parameters: NDArray) -> NDArray:
    """
    Returns the triggers for this propagator.
    :param n: the number of variables
    :param parameters: the parameters, unused here
    :return: an array of triggers
    """
    triggers = np.full(n, dtype=np.uint8, fill_value=EVENT_MASK_MAX)
    triggers[-1] = EVENT_MASK_MIN
    return triggers


@njit(cache=True)
def compute_domains_min_geq(domains: NDArray, parameters: NDArray) -> int:
    """
    Implements Min_i x_i >= x_{n-1}.
    :param domains: the domains of the variables, x is an alias for domains
    :param parameters: unused here
    :return: the status of the propagation (consistency, inconsistency or entailment) as an int
    """
    x = domains[:-1]
    y = domains[-1]
    if y[MAX] <= np.min(x[:, MIN]):
        return PROP_ENTAILMENT
    y[MAX] = min(y[MAX], np.min(x[:, MAX]))
    if y[MIN] > y[MAX]:
        return PROP_INCONSISTENCY
    for i in range(len(x)):
        x[i, MIN] = max(x[i, MIN], y[MIN])
        if x[i, MAX] < x[i, MIN]:
            return PROP_INCONSISTENCY
    return PROP_CONSISTENCY



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
from typing import List

from nucs.problems.problem import Problem
from nucs.propagators.propagators import ALG_AFFINE_EQ, ALG_AFFINE_LEQ, ALG_ALLDIFFERENT


class MagicSquareProblem(Problem):
    """
    A simple model for magic squares.

    CSPLIB problem #19 - https://www.csplib.org/Problems/prob019/
    """

    def __init__(self, n: int, symmetry_breaking: bool = True):
        """
        Inits the problem.
        :param n: the size of the square
        :param symmetry_breaking: a boolean indicating if symmetry constraints should be added to the model
        """
        self.n = n
        self.m = (n**2 - 1) * n / 2
        super().__init__([(0, n**2 - 1)] * n**2)
        for i in range(n):
            self.add_propagator((self.row(i), ALG_AFFINE_EQ, [1] * n + [self.m]))
            self.add_propagator((self.column(i), ALG_AFFINE_EQ, [1] * n + [self.m]))
        self.add_propagator((self.first_diag(), ALG_AFFINE_EQ, [1] * n + [self.m]))
        self.add_propagator((self.second_diag(), ALG_AFFINE_EQ, [1] * n + [self.m]))
        self.add_propagator((list(range(n**2)), ALG_ALLDIFFERENT, []))
        if symmetry_breaking:
            top_left = self.first_diag()[0]
            bottom_right = self.first_diag()[-1]
            top_right = self.second_diag()[0]
            bottom_left = self.second_diag()[-1]
            self.add_propagator(([top_left, top_right], ALG_AFFINE_LEQ, [1, -1, -1]))
            self.add_propagator(([top_left, bottom_left], ALG_AFFINE_LEQ, [1, -1, -1]))
            self.add_propagator(([top_left, bottom_right], ALG_AFFINE_LEQ, [1, -1, -1]))
            self.add_propagator(([top_right, bottom_left], ALG_AFFINE_LEQ, [1, -1, -1]))

    def row(self, i: int) -> List[int]:
        return list(range(0 + i * self.n, self.n + i * self.n))

    def column(self, j: int) -> List[int]:
        return list(range(j, self.n**2, self.n))

    def first_diag(self) -> List[int]:
        return list(range(0, self.n**2, self.n + 1))

    def second_diag(self) -> List[int]:
        return list(range(self.n**2 - self.n, 0, 1 - self.n))



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
from typing import List, Tuple, Union

import numpy as np
from numpy.typing import NDArray


def new_shr_domains_by_values(domains: List[Union[int, Tuple[int, int]]]) -> NDArray:
    """
    Only used in tests.
    """
    return np.array(
        [(domain, domain) if isinstance(domain, int) else domain for domain in domains], dtype=np.int32
    )  # , order="F"


def new_parameters_by_values(data: List[int]) -> NDArray:
    """
    Only used in tests.
    """
    return np.array(data, dtype=np.int32)



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
from typing import List

from numpy.typing import NDArray

from nucs.problems.problem import Problem
from nucs.propagators.propagators import ALG_ALLDIFFERENT, ALG_EXACTLY_EQ, ALG_GCC, ALG_RELATION


class SportsTournamentSchedulingProblem(Problem):
    """
    The problem is to schedule a tournament of n teams over n−1 weeks,
    with each week divided into n/2 periods,
    and each period divided into two slots.

    The first team in each slot plays at home, whilst the second plays the first team away.

    A tournament must satisfy the following three constraints:
    - every team plays once a week;
    - every team plays at most twice in the same period over the tournament;
    - every team plays every other team.

    CSPLIB problem #26 - https://www.csplib.org/Problems/prob026/
    """

    def solution_as_matrix(self, solution: NDArray) -> List[List[str]]:
        """
        Returns the solutions as a matrix of strings.
        :param solution: the solution as a list of ints
        :return: a matrix
        """
        return [
            [
                f"{solution[self.team_var_index(p, w, 0)]}-{solution[self.team_var_index(p, w, 1)]}"
                for w in range(0, self.week_nb)
            ]
            for p in range(0, self.period_nb)
        ]

    def team_var_index(self, p: int, w: int, s: int) -> int:
        return p * (self.week_nb * self.slot_nb) + w * self.slot_nb + s

    def match_var_index(self, p: int, w: int) -> int:
        return self.team_var_nb + p * self.week_nb + w

    def teams_per_week(self, w: int) -> List[int]:
        return [self.team_var_index(p, w, s) for p in range(self.period_nb) for s in range(self.slot_nb)]

    def teams_per_period(self, p: int) -> List[int]:
        return [self.team_var_index(p, w, s) for w in range(self.week_nb) for s in range(self.slot_nb)]

    def matches(self) -> List[int]:
        return list(range(self.team_var_nb, self.team_var_nb + self.match_nb))

    def matches_per_week(self, w: int) -> List[int]:
        return [self.match_var_index(p, w) for p in range(self.period_nb)]

    def match_ordinal(self, t1: int, t2: int) -> int:
        return self.match_nb - ((self.team_nb - t1) * (self.team_nb - t1 - 1)) // 2 + t2 - t1 - 1

    def plays(self) -> List[int]:
        plays = []
        for i in range(0, self.team_nb - 1):
            for j in range(i + 1, self.team_nb):
                plays.extend([i, j, self.match_ordinal(i, j)])
        return plays

    def __init__(self, n: int, symmetry_breaking: bool = True) -> None:
        self.team_nb = n
        self.slot_nb = 2
        self.period_nb = n // 2
        self.week_nb = n - 1
        self.match_nb = ((n - 1) * n) // 2
        self.team_var_nb = self.period_nb * self.week_nb * self.slot_nb
        super().__init__([(0, self.team_nb - 1)] * self.team_var_nb + [(0, self.match_nb - 1)] * self.match_nb)
        plays = self.plays()
        self.add_propagator((self.matches(), ALG_ALLDIFFERENT, []))  # matches are different
        self.add_propagators([(self.teams_per_week(w), ALG_ALLDIFFERENT, []) for w in range(0, self.week_nb)])
        self.add_propagators(
            [(self.teams_per_period(p), ALG_GCC, [0] + [1] * n + [2] * n) for p in range(0, self.period_nb)]
        )
        self.add_propagators(
            [
                (
                    [self.team_var_index(p, w, 0), self.team_var_index(p, w, 1), self.match_var_index(p, w)],
                    ALG_RELATION,
                    plays,
                )
                for p in range(0, self.period_nb)
                for w in range(0, self.week_nb)
            ]
        )
        if symmetry_breaking:
            # the first week is set
            k = 0
            for p in range(self.period_nb):
                for s in range(self.slot_nb):
                    self.shr_domains_lst[self.team_var_index(p, 0, s)] = [k, k]
                    k += 1
            # the match `0 versus (t+1)` appears at week t (much slower)
            self.add_propagators(
                [
                    (self.matches_per_week(w), ALG_EXACTLY_EQ, [self.match_ordinal(0, w + 1), 1])
                    for w in range(self.week_nb)
                ]
            )



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
from nucs.examples.sports_tournament_scheduling.sports_tournament_scheduling_problem import (
    SportsTournamentSchedulingProblem,
)
from nucs.heuristics.heuristics import DOM_HEURISTIC_MIN_VALUE, VAR_HEURISTIC_SMALLEST_DOMAIN
from nucs.solvers.backtrack_solver import BacktrackSolver


class TestSportsTournamentScheduling:
    def test_teams_per_week(self) -> None:
        problem = SportsTournamentSchedulingProblem(4)
        assert problem.teams_per_week(0) == [0, 1, 6, 7]
        assert problem.teams_per_week(1) == [2, 3, 8, 9]
        assert problem.teams_per_week(2) == [4, 5, 10, 11]

    def test_teams_per_period(self) -> None:
        problem = SportsTournamentSchedulingProblem(4)
        assert problem.teams_per_period(0) == [0, 1, 2, 3, 4, 5]
        assert problem.teams_per_period(1) == [6, 7, 8, 9, 10, 11]

    def test_plays(self) -> None:
        problem = SportsTournamentSchedulingProblem(4)
        assert problem.plays() == [0, 1, 0, 0, 2, 1, 0, 3, 2, 1, 2, 3, 1, 3, 4, 2, 3, 5]

    def test_sports_tournament_scheduling_check(self) -> None:
        problem = SportsTournamentSchedulingProblem(8, False)
        problem.shr_domains_lst[:56] = [
            [v, v]
            for v in [
                0,
                1,
                0,
                2,
                4,
                7,
                3,
                6,
                3,
                7,
                1,
                5,
                2,
                4,
                2,
                3,
                1,
                7,
                0,
                3,
                5,
                7,
                1,
                4,
                0,
                6,
                5,
                6,
                4,
                5,
                3,
                5,
                1,
                6,
                0,
                4,
                2,
                6,
                2,
                7,
                0,
                7,
                6,
                7,
                4,
                6,
                2,
                5,
                1,
                2,
                0,
                5,
                3,
                4,
                1,
                3,
            ]
        ]
        solver = BacktrackSolver(problem)
        assert next(solver.solve()) is not None

    def test_sports_tournament_scheduling_solve(self) -> None:
        problem = SportsTournamentSchedulingProblem(8)
        solver = BacktrackSolver(
            problem, var_heuristic_idx=VAR_HEURISTIC_SMALLEST_DOMAIN, dom_heuristic_idx=DOM_HEURISTIC_MIN_VALUE
        )
        assert next(solver.solve()) is not None



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
import math

import numpy as np
from numba import njit  # type: ignore
from numpy.typing import NDArray

from nucs.constants import EVENT_MASK_MIN_MAX, MAX, MIN, PROP_CONSISTENCY, PROP_INCONSISTENCY


def get_complexity_alldifferent(n: int, parameters: NDArray) -> float:
    """
    Returns the time complexity of the propagator as a float.
    :param n: the number of variables
    :param parameters: the parameters, unused here
    :return: a float
    """
    return 2 * n * math.log(n) + 5 * n


def get_triggers_alldifferent(n: int, parameters: NDArray) -> NDArray:
    """
    This propagator is triggered whenever there is a change in the domain of a variable.
    :param n: the number of variables
    :param parameters: the parameters, unused here
    :return: an array of triggers
    """
    return np.full(n, dtype=np.uint8, fill_value=EVENT_MASK_MIN_MAX)


@njit(cache=True)
def path_set(t: NDArray, start: int, end: int, value: int) -> None:
    """
    Sets t[start], t[t[start]], ..., a[idx] to value until a[idx] = end.
    :param t: an array of pointers
    :param start: an index
    :param end: an index
    :param value: a value
    """
    while (p := start) != end:
        start = t[p]
        t[p] = value


@njit(cache=True)
def path_min(t: NDArray, i: int) -> int:
    """
    Follows i, t[i], t[t[i], ... until it stops decreasing.
    :param t: an array of pointers
    :param i: an index
    :return: the index found
    """
    while t[i] < i:
        i = t[i]
    return i


@njit(cache=True)
def path_max(t: NDArray, i: int) -> int:
    """
    Follows i, t[i], t[t[i], ... until it stops increasing.
    :param t: an array of pointers
    :param i: an index
    :return: the index found
    """
    while t[i] > i:
        i = t[i]
    return i


@njit(cache=True)
def update_bounds(
    bounds: NDArray,
    n: int,
    domains: NDArray,
    ranks: NDArray,
    min_sorted_vars: NDArray,
    max_sorted_vars: NDArray,
) -> int:
    min_value = domains[min_sorted_vars[0], MIN]
    max_value = domains[max_sorted_vars[0], MAX] + 1
    last = min_value - 2
    bounds[0] = last
    i = j = nb = 0
    while True:
        if i < n and min_value <= max_value:
            if min_value != last:
                nb += 1
                bounds[nb] = last = min_value
            ranks[min_sorted_vars[i], MIN] = nb
            i += 1
            if i < n:
                min_value = domains[min_sorted_vars[i], MIN]
        else:
            if max_value != last:
                nb += 1
                bounds[nb] = last = max_value
            ranks[max_sorted_vars[j], MAX] = nb
            j += 1
            if j == n:
                break
            max_value = domains[max_sorted_vars[j], MAX] + 1
    bounds[nb + 1] = bounds[nb] + 2
    return nb


@njit(cache=True)
def filter_lower(
    n: int,
    nb: int,
    t: NDArray,
    d: NDArray,
    h: NDArray,
    bounds: NDArray,
    domains: NDArray,
    ranks: NDArray,
    max_sorted_vars: NDArray,
) -> bool:
    for i in range(1, nb + 2):
        t[i] = h[i] = i - 1
        d[i] = bounds[i] - bounds[i - 1]
    for i, max_sorted_vars_i in enumerate(max_sorted_vars):
        x = ranks[max_sorted_vars_i, MIN]
        y = ranks[max_sorted_vars_i, MAX]
        z = path_max(t, x + 1)
        j = t[z]
        d[z] -= 1
        if d[z] == 0:
            t[z] = z + 1
            z = path_max(t, t[z])
            t[z] = j
        if d[z] + bounds[y] < bounds[z]:  # moved above the path compression which is not the case in the paper
            return False
        path_set(t, x + 1, z, z)  # path compression
        if h[x] > x:
            w = path_max(h, h[x])
            domains[max_sorted_vars_i, MIN] = bounds[w]
            path_set(h, x, w, w)  # path compression
        if d[z] + bounds[y] == bounds[z]:
            path_set(h, h[y], j - 1, y)  # mark hall interval
            h[y] = j - 1  # hall interval[bounds[j], bounds[y]]
    return True


@njit(cache=True)
def filter_upper(
    n: int,
    nb: int,
    t: NDArray,
    d: NDArray,
    h: NDArray,
    bounds: NDArray,
    domains: NDArray,
    ranks: NDArray,
    min_sorted_vars: NDArray,
) -> bool:
    for i in range(nb + 1):
        t[i] = h[i] = i + 1
        d[i] = bounds[i + 1] - bounds[i]
    for i in range(n - 1, -1, -1):
        min_sorted_vars_i = min_sorted_vars[i]
        x = ranks[min_sorted_vars_i, MAX]
        y = ranks[min_sorted_vars_i, MIN]
        z = path_min(t, x - 1)
        j = t[z]
        d[z] -= 1
        if d[z] == 0:
            t[z] = z - 1
            z = path_min(t, t[z])
            t[z] = j
        if d[z] + bounds[z] < bounds[y]:  # moved above the path compression which is not the case in the paper
            return False
        path_set(t, x - 1, z, z)  # path compression
        if h[x] < x:
            w = path_min(h, h[x])
            domains[min_sorted_vars_i, MAX] = bounds[w] - 1
            path_set(h, x, w, w)  # path compression
        if d[z] + bounds[z] == bounds[y]:
            path_set(h, h[y], j + 1, y)  # mark hall interval
            h[y] = j + 1  # hall interval[bounds[j], bounds[y]]
    return True


@njit(cache=True)
def compute_domains_alldifferent(domains: NDArray, parameters: NDArray) -> int:
    """
    Enforces that x_i <> x_j when i<>j.
    Adapted from "A fast and simple algorithm for bounds consistency of the alldifferent constraint".
    :param domains: the domains of the variables, x is an alias for domains
    :param parameters: unused here
    :return: the status of the propagation (consistency, inconsistency or entailment) as an int
    """
    n = len(domains)
    ranks = np.zeros((n, 2), dtype=np.uint16)
    bounds_nb = 2 * n + 2
    bounds = np.zeros(bounds_nb, dtype=np.int32)
    t = np.zeros(bounds_nb, dtype=np.uint16)  # critical capacity pointers
    d = np.zeros(bounds_nb, dtype=np.int32)  # differences between critical capacities
    h = np.zeros(bounds_nb, dtype=np.uint16)  # Hall interval pointers
    min_sorted_vars = np.argsort(domains[:, MIN])
    max_sorted_vars = np.argsort(domains[:, MAX])
    nb = update_bounds(bounds, n, domains, ranks, min_sorted_vars, max_sorted_vars)
    return (
        PROP_CONSISTENCY
        if filter_lower(n, nb, t, d, h, bounds, domains, ranks, max_sorted_vars)
        and filter_upper(n, nb, t, d, h, bounds, domains, ranks, min_sorted_vars)
        else PROP_INCONSISTENCY
    )



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
from typing import Callable

from nucs.heuristics.first_not_instantiated_var_heuristic import first_not_instantiated_var_heuristic
from nucs.heuristics.greatest_domain_var_heuristic import greatest_domain_var_heuristic
from nucs.heuristics.max_regret_var_heuristic import max_regret_var_heuristic
from nucs.heuristics.max_value_dom_heuristic import max_value_dom_heuristic
from nucs.heuristics.mid_value_dom_heuristic import mid_value_dom_heuristic
from nucs.heuristics.min_cost_dom_heuristic import min_cost_dom_heuristic
from nucs.heuristics.min_value_dom_heuristic import min_value_dom_heuristic
from nucs.heuristics.smallest_domain_var_heuristic import smallest_domain_var_heuristic
from nucs.heuristics.split_high_dom_heuristic import split_high_dom_heuristic
from nucs.heuristics.split_low_dom_heuristic import split_low_dom_heuristic
from nucs.heuristics.split_random_dom_heuristic import split_random_dom_heuristic

VAR_HEURISTIC_FCTS = []
DOM_HEURISTIC_FCTS = []


def register_var_heuristic(var_heuristic_fct: Callable) -> int:
    """
    Register a variable heuristic by adding it function to the corresponding list of functions.
    :param var_heuristic_fct: a function that implements the variable heuristic
    :return: the index of the variable heuristic
    """
    VAR_HEURISTIC_FCTS.append(var_heuristic_fct)
    return len(VAR_HEURISTIC_FCTS) - 1


def register_dom_heuristic(dom_heuristic_fct: Callable) -> int:
    """
    Register a domain heuristic by adding it function to the corresponding list of functions.
    :param dom_heuristic_fct: a function that implements the domain heuristic
    :return: the index of the domain heuristic
    """
    DOM_HEURISTIC_FCTS.append(dom_heuristic_fct)
    return len(DOM_HEURISTIC_FCTS) - 1


VAR_HEURISTIC_FIRST_NOT_INSTANTIATED = register_var_heuristic(first_not_instantiated_var_heuristic)
VAR_HEURISTIC_GREATEST_DOMAIN = register_var_heuristic(greatest_domain_var_heuristic)
VAR_HEURISTIC_MAX_REGRET = register_var_heuristic(max_regret_var_heuristic)
VAR_HEURISTIC_SMALLEST_DOMAIN = register_var_heuristic(smallest_domain_var_heuristic)

DOM_HEURISTIC_MAX_VALUE = register_dom_heuristic(max_value_dom_heuristic)
DOM_HEURISTIC_MID_VALUE = register_dom_heuristic(mid_value_dom_heuristic)
DOM_HEURISTIC_MIN_COST = register_dom_heuristic(min_cost_dom_heuristic)
DOM_HEURISTIC_MIN_VALUE = register_dom_heuristic(min_value_dom_heuristic)
DOM_HEURISTIC_SPLIT_HIGH = register_dom_heuristic(split_high_dom_heuristic)
DOM_HEURISTIC_SPLIT_LOW = register_dom_heuristic(split_low_dom_heuristic)
DOM_HEURISTIC_SPLIT_RANDOM = register_dom_heuristic(split_random_dom_heuristic)



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
import numpy as np
from numba import njit  # type: ignore
from numpy.typing import NDArray

from nucs.constants import EVENT_MASK_MIN_MAX, MAX, MIN, PROP_CONSISTENCY, PROP_ENTAILMENT, PROP_INCONSISTENCY


def get_complexity_exactly_eq(n: int, parameters: NDArray) -> float:
    """
    Returns the time complexity of the propagator as a float.
    :param n: the number of variables
    :param parameters: the parameters, unused here
    :return: a float
    """
    return 2 * n


def get_triggers_exactly_eq(n: int, parameters: NDArray) -> NDArray:
    """
    This propagator is triggered whenever there is a change in the domain of a variable.
    :param n: the number of variables
    :param parameters: the parameters, unused here
    :return: an array of triggers
    """
    return np.full(n, dtype=np.uint8, fill_value=EVENT_MASK_MIN_MAX)


@njit(cache=True)
def compute_domains_exactly_eq(domains: NDArray, parameters: NDArray) -> int:
    """
    Implements Sigma_i (x_i == a) = c.
    :param domains: the domains of the variables, x is an alias for domains
    :param parameters: the parameters of the propagator, a is the first parameter, c is the second parameter
    :return: the status of the propagation (consistency, inconsistency or entailment) as an int
    """
    a = parameters[0]
    c = parameters[1]
    count_max = len(domains) - c
    count_min = -c
    for domain in domains:
        if domain[MIN] > a or domain[MAX] < a:
            count_max -= 1
            if count_max < 0:
                return PROP_INCONSISTENCY
        elif domain[MIN] == a and domain[MAX] == a:
            count_min += 1
            if count_min > 0:
                return PROP_INCONSISTENCY
    if count_min == 0 and count_max == 0:
        return PROP_ENTAILMENT
    if count_min == 0:  # we cannot have more domains equal to a
        for domain in domains:
            if domain[MIN] == a and domain[MAX] > a:
                domain[MIN] = a + 1
            if domain[MIN] < a and domain[MAX] == a:
                domain[MAX] = a - 1
    elif count_max == 0:  # we cannot have more domains different from a
        for domain in domains:
            if domain[MIN] <= a <= domain[MAX]:
                domain[:] = a
    return PROP_CONSISTENCY



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
import random

from numba import njit  # type: ignore
from numpy.typing import NDArray

from nucs.heuristics.split_high_dom_heuristic import split_high_dom_heuristic
from nucs.heuristics.split_low_dom_heuristic import split_low_dom_heuristic


@njit(cache=True)
def split_random_dom_heuristic(
    shr_domains_stack: NDArray,
    not_entailed_propagators_stack: NDArray,
    dom_update_stack: NDArray,
    stacks_top: NDArray,
    dom_idx: int,
    params: NDArray,
) -> int:
    if random.randint(0, 1) == 0:
        return split_low_dom_heuristic(
            shr_domains_stack, not_entailed_propagators_stack, dom_update_stack, stacks_top, dom_idx, params
        )
    else:
        return split_high_dom_heuristic(
            shr_domains_stack, not_entailed_propagators_stack, dom_update_stack, stacks_top, dom_idx, params
        )



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
import argparse

from rich import print

from nucs.constants import LOG_LEVEL_INFO, LOG_LEVELS
from nucs.examples.quasigroup.quasigroup_problem import QuasigroupProblem
from nucs.heuristics.heuristics import DOM_HEURISTIC_SPLIT_LOW, VAR_HEURISTIC_SMALLEST_DOMAIN
from nucs.solvers.backtrack_solver import BacktrackSolver
from nucs.solvers.consistency_algorithms import CONSISTENCY_ALG_BC, CONSISTENCY_ALG_SHAVING
from nucs.solvers.multiprocessing_solver import MultiprocessingSolver

# Run with the following command (the second run is much faster because the code has been compiled):
# NUMBA_CACHE_DIR=.numba/cache python -m nucs.examples.quasigroup -n 10 --symmetry_breaking
if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--log_level", choices=LOG_LEVELS, default=LOG_LEVEL_INFO)
    parser.add_argument("-n", type=int, default=10)
    parser.add_argument("--idempotent", action=argparse.BooleanOptionalAction, default=True)
    parser.add_argument("--shaving", type=bool, action=argparse.BooleanOptionalAction, default=False)
    parser.add_argument("--symmetry_breaking", action=argparse.BooleanOptionalAction, default=True)
    parser.add_argument("--processors", type=int, default=1)
    parser.add_argument("--kind", type=int, choices=[3, 4, 5, 6, 7], default=5)
    args = parser.parse_args()
    problem = QuasigroupProblem(args.kind, args.n, args.idempotent, args.symmetry_breaking)
    solver = (
        MultiprocessingSolver(
            [
                BacktrackSolver(
                    problem,
                    decision_domains=list(range(0, args.n * args.n)),
                    consistency_alg_idx=CONSISTENCY_ALG_SHAVING if args.shaving else CONSISTENCY_ALG_BC,
                    var_heuristic_idx=VAR_HEURISTIC_SMALLEST_DOMAIN,
                    dom_heuristic_idx=DOM_HEURISTIC_SPLIT_LOW,
                    log_level=args.log_level,
                )
                for problem in problem.split(args.processors, 1)
            ]
        )
        if args.processors > 1
        else BacktrackSolver(
            problem,
            decision_domains=list(range(0, args.n * args.n)),
            consistency_alg_idx=CONSISTENCY_ALG_SHAVING if args.shaving else CONSISTENCY_ALG_BC,
            var_heuristic_idx=VAR_HEURISTIC_SMALLEST_DOMAIN,
            dom_heuristic_idx=DOM_HEURISTIC_SPLIT_LOW,
            log_level=args.log_level,
        )
    )
    solver.find_one()
    print(solver.get_statistics())



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
import argparse

from rich import print

from nucs.constants import LOG_LEVEL_INFO, LOG_LEVELS, OPTIM_MODES, OPTIM_PRUNE
from nucs.examples.tsp.tsp_instances import TSP_INSTANCES
from nucs.examples.tsp.tsp_problem import TSPProblem
from nucs.examples.tsp.tsp_var_heuristic import tsp_var_heuristic
from nucs.heuristics.heuristics import DOM_HEURISTIC_MIN_COST, register_var_heuristic
from nucs.solvers.backtrack_solver import BacktrackSolver
from nucs.solvers.consistency_algorithms import CONSISTENCY_ALG_BC, CONSISTENCY_ALG_SHAVING
from nucs.solvers.multiprocessing_solver import MultiprocessingSolver

# Run with the following command (the second run is much faster because the code has been compiled):
# NUMBA_CACHE_DIR=.numba/cache python -m nucs.examples.alpha
if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--log_level", choices=LOG_LEVELS, default=LOG_LEVEL_INFO)
    parser.add_argument("--name", choices=["GR17", "GR21", "GR24"], default="GR17")
    parser.add_argument("--opt_mode", choices=OPTIM_MODES, default=OPTIM_PRUNE)
    parser.add_argument("--processors", type=int, default=1)
    parser.add_argument("--shaving", type=bool, action=argparse.BooleanOptionalAction, default=False)
    args = parser.parse_args()
    tsp_instance = TSP_INSTANCES[args.name]
    n = len(tsp_instance)
    decision_domains = list(range(0, 2 * n))
    problem = TSPProblem(tsp_instance)
    costs = tsp_instance + tsp_instance  # symmetry of costs
    tsp_var_heuristic_idx = register_var_heuristic(tsp_var_heuristic)
    solver = (
        MultiprocessingSolver(
            [
                BacktrackSolver(
                    prob,
                    consistency_alg_idx=CONSISTENCY_ALG_SHAVING if args.shaving else CONSISTENCY_ALG_BC,
                    decision_domains=decision_domains,
                    var_heuristic_idx=tsp_var_heuristic_idx,
                    var_heuristic_params=costs,
                    dom_heuristic_idx=DOM_HEURISTIC_MIN_COST,
                    dom_heuristic_params=costs,
                    log_level=args.log_level,
                )
                for prob in problem.split(args.processors, 0)
            ]
        )
        if args.processors > 1
        else BacktrackSolver(
            problem,
            consistency_alg_idx=CONSISTENCY_ALG_SHAVING if args.shaving else CONSISTENCY_ALG_BC,
            decision_domains=decision_domains,
            var_heuristic_idx=tsp_var_heuristic_idx,
            var_heuristic_params=costs,
            dom_heuristic_idx=DOM_HEURISTIC_MIN_COST,
            dom_heuristic_params=costs,
            log_level=args.log_level,
        )
    )
    solution = solver.minimize(problem.total_cost, mode=args.opt_mode)
    print(solver.get_statistics())



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
import numpy as np
from numba import njit  # type: ignore
from numpy.typing import NDArray

from nucs.constants import EVENT_MASK_MIN_MAX, MAX, MIN, PROP_CONSISTENCY, PROP_INCONSISTENCY


def get_complexity_permutation_aux(n: int, parameters: NDArray) -> float:
    """
    Returns the time complexity of the propagator as a float.
    :param n: the number of variables
    :param parameters: the parameters, unused here
    :return: a float
    """
    return n * n


def get_triggers_permutation_aux(n: int, parameters: NDArray) -> NDArray:
    """
    :param n: the number of variables
    :param parameters: the parameters, unused here
    :return: an array of triggers
    """
    return np.full(n, dtype=np.uint8, fill_value=EVENT_MASK_MIN_MAX)


@njit(cache=True)
def compute_domains_permutation_aux(domains: NDArray, parameters: NDArray) -> int:
    """
    An auxiliary propagator needed to connect the next and prev variables of a permutation problem.
    :param domains: the domains of the variables, the next and prev variables
    :param parameters: the parameters of the propagator, unused here
    :return: the status of the propagation (consistency, inconsistency or entailment) as an int
    """
    n = len(domains) // 2
    next = domains[:n]
    prev = domains[n:]
    return (
        PROP_CONSISTENCY
        if filter_domains_permutation(n, next, prev) and filter_domains_permutation(n, prev, next)
        else PROP_INCONSISTENCY
    )


@njit(cache=True)
def filter_domains_permutation(n: int, next: NDArray, prev: NDArray) -> bool:
    for j in range(0, n):
        if prev[j, MIN] == prev[j, MAX]:
            i = prev[j, MIN]
            next[i, :] = j
        else:
            start = -1
            for i in range(prev[j, MIN], prev[j, MAX] + 1):
                if j < next[i, MIN] or j > next[i, MAX]:  # then prev[j] != i
                    if start == -1:
                        start = i
                    if i == prev[j, MIN]:
                        prev[j, MIN] += 1
                else:
                    start = -1
            if start >= 0:
                prev[j, MAX] = start - 1
                if prev[j, MAX] < prev[j, MIN]:
                    return False
    return True



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
from nucs.problems.latin_square_problem import M_COLOR, M_COLUMN, M_ROW, LatinSquareRCProblem
from nucs.propagators.propagators import ALG_ELEMENT_LIV_ALLDIFFERENT


class QuasigroupProblem(LatinSquareRCProblem):
    """
    CSPLIB problem #3 - https://www.csplib.org/Problems/prob003/
    """

    def __init__(self, kind: int, n: int, idempotent: bool, symmetry_breaking: bool = True):
        """
        Inits the problem.
        :param n: the size of the quasigroup
        :param idempotent: a boolean indicating if the quasigroup is idempotent
        :param symmetry_breaking: a boolean indicating if symmetry constraints should be added to the model
        """
        super().__init__(n)
        if symmetry_breaking:
            for i in range(1, n):
                self.shr_domains_lst[self.cell(i, n - 1)] = [i - 1, n - 1]
        if idempotent:
            for model in [M_COLOR, M_ROW, M_COLUMN]:
                for i in range(n):
                    self.shr_domains_lst[self.cell(i, i, model)] = [i, i]
                for i in range(1, n):
                    self.shr_domains_lst[self.cell(0, i, model)] = [1, n - 1]
                for i in range(0, n - 1):
                    self.shr_domains_lst[self.cell(n - 1, i, model)] = [0, n - 2]
        for i in range(n):
            for j in range(n):
                if not idempotent or j != i:
                    if kind == 3:  # (a∗b)∗(b*a)=a
                        # color[color[i, j], color[j, i]] = i
                        # column[color[i, j], i] = color[j, i] which avoids the creation of additional variables
                        self.add_propagator(
                            (
                                [*self.column(i, M_COLUMN), self.cell(i, j), self.cell(j, i)],
                                ALG_ELEMENT_LIV_ALLDIFFERENT,
                                [],
                            )
                        )
                    elif kind == 4:  # (b∗a)∗(a*b) = a
                        # color[color[j, i], color[i, j]] = i
                        # column[color[j, i], i] = color[i, j] which avoids the creation of additional variables
                        self.add_propagator(
                            (
                                [*self.column(i, M_COLUMN), self.cell(j, i), self.cell(i, j)],
                                ALG_ELEMENT_LIV_ALLDIFFERENT,
                                [],
                            )
                        )
                    elif kind == 5:  # ((b∗a)∗b)∗b = a
                        # color[color[color[j, i], j], j] = i
                        # row[i, j] = color[color[j, i], j] which avoids the creation of additional variables
                        self.add_propagator(
                            (
                                [*self.column(j), self.cell(j, i), self.cell(i, j, M_ROW)],
                                ALG_ELEMENT_LIV_ALLDIFFERENT,
                                [],
                            )
                        )
                    elif kind == 6:  # (a∗b)∗b = a*(a*b)
                        # color[color[i, j], j] = color[i, color[i, j]]
                        ijj = self.add_variable((0, n - 1))
                        self.add_propagator(([*self.column(j), self.cell(i, j), ijj], ALG_ELEMENT_LIV_ALLDIFFERENT, []))
                        self.add_propagator(([*self.row(i), self.cell(i, j), ijj], ALG_ELEMENT_LIV_ALLDIFFERENT, []))
                    elif kind == 7:  # (b∗a)∗b = a*(b*a)
                        # color[color[j, i], j] = color[i, color[j, i]]
                        jij = self.add_variable((0, n - 1))
                        self.add_propagator(([*self.column(j), self.cell(j, i), jij], ALG_ELEMENT_LIV_ALLDIFFERENT, []))
                        self.add_propagator(([*self.row(i), self.cell(j, i), jij], ALG_ELEMENT_LIV_ALLDIFFERENT, []))



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
import numpy as np

from nucs.constants import PROP_CONSISTENCY, PROP_ENTAILMENT, PROP_INCONSISTENCY
from nucs.numpy_helper import new_parameters_by_values, new_shr_domains_by_values
from nucs.propagators.element_lic_propagator import compute_domains_element_lic


class TestElementLIC:
    def test_compute_domains_1(self) -> None:
        domains = new_shr_domains_by_values([(-1, 1), (1, 2), (0, 2)])
        data = new_parameters_by_values([1])
        assert compute_domains_element_lic(domains, data) == PROP_CONSISTENCY
        assert np.all(domains == np.array([[-1, 1], [1, 2], [0, 1]]))

    def test_compute_domains_2(self) -> None:
        domains = new_shr_domains_by_values([(-4, -2), (1, 2), (0, 1)])
        data = new_parameters_by_values([1])
        assert compute_domains_element_lic(domains, data) == PROP_ENTAILMENT
        assert np.all(domains == np.array([[-4, -2], [1, 1], [1, 1]]))

    def test_compute_domains_3(self) -> None:
        domains = new_shr_domains_by_values([(-4, -2), (1, 2), (0, 1)])
        data = new_parameters_by_values([0])
        assert compute_domains_element_lic(domains, data) == PROP_INCONSISTENCY



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
from numba import njit  # type: ignore
from numpy.typing import NDArray

from nucs.constants import MAX, MIN
from nucs.heuristics.value_dom_heuristic import value_dom_heuristic


@njit(cache=True)
def mid_value_dom_heuristic(
    shr_domains_stack: NDArray,
    not_entailed_propagators_stack: NDArray,
    dom_update_stack: NDArray,
    stacks_top: NDArray,
    dom_idx: int,
    params: NDArray,
) -> int:
    """
    Chooses the middle value of the domain.
    :param shr_domains_stack: the stack of shared domains
    :param not_entailed_propagators_stack: the stack of not entailed propagators
    :param dom_update_stack: the stack of domain updates
    :param stacks_top: the index of the top of the stacks as a Numpy array
    :param dom_idx: the index of the shared domain
    :param params: a two-dimensional parameters array, unused here
    :return: the events
    """
    return value_dom_heuristic(
        shr_domains_stack,
        not_entailed_propagators_stack,
        dom_update_stack,
        stacks_top,
        dom_idx,
        (shr_domains_stack[stacks_top[0], dom_idx, MIN] + shr_domains_stack[stacks_top[0], dom_idx, MAX]) // 2,
        params,
    )



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
import pytest

from nucs.constants import STATS_IDX_SOLVER_SOLUTION_NB
from nucs.problems.latin_square_problem import LatinSquareProblem, LatinSquareRCProblem
from nucs.solvers.backtrack_solver import BacktrackSolver


class TestLatinSquare:
    @pytest.mark.parametrize(
        "size, solution_nb",
        [
            (1, 1),
            (2, 2),
            (3, 12),
            (4, 576),
            # (5, 161280)
        ],
    )
    def test_latin_square(self, size: int, solution_nb: int) -> None:
        problem = LatinSquareProblem(list(range(size)))
        solver = BacktrackSolver(problem)
        solver.solve_all()
        assert solver.statistics[STATS_IDX_SOLVER_SOLUTION_NB] == solution_nb

    @pytest.mark.parametrize(
        "size, solution_nb",
        [
            (1, 1),
            (2, 2),
            (3, 12),
            (4, 576),
            # (5, 161280)
        ],
    )
    def test_latin_square_rc(self, size: int, solution_nb: int) -> None:
        problem = LatinSquareRCProblem(size)
        solver = BacktrackSolver(problem)
        solver.solve_all()
        assert solver.statistics[STATS_IDX_SOLVER_SOLUTION_NB] == solution_nb



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
import numpy as np
from numba import njit  # type: ignore
from numpy.typing import NDArray

from nucs.constants import EVENT_MASK_MIN_MAX, MAX, MIN, PROP_CONSISTENCY, PROP_ENTAILMENT, PROP_INCONSISTENCY


def get_complexity_count_eq(n: int, parameters: NDArray) -> float:
    """
    Returns the time complexity of the propagator as a float.
    :param n: the number of variables
    :param parameters: the parameters, unused here
    :return: a float
    """
    return 2 * n


def get_triggers_count_eq(n: int, parameters: NDArray) -> NDArray:
    """
    This propagator is triggered whenever there is a change in the domain of a variable.
    :param n: the number of variables
    :param parameters: the parameters, unused here
    :return: an array of triggers
    """
    return np.full(n, dtype=np.uint8, fill_value=EVENT_MASK_MIN_MAX)


@njit(cache=True)
def compute_domains_count_eq(domains: NDArray, parameters: NDArray) -> int:
    """
    Implements Sigma_i (x_i == a) = x_{n-1}.
    :param domains: the domains of the variables, x is an alias for domains
    :param parameters: the parameters of the propagator, a is the first parameter
    :return: the status of the propagation (consistency, inconsistency or entailment) as an int
    """
    a = parameters[0]
    x = domains[:-1]
    counter = domains[-1]
    count_max = len(x)
    count_min = 0
    for x_i in x:
        if x_i[MIN] > a or x_i[MAX] < a:
            count_max -= 1
        elif x_i[MIN] == a and x_i[MAX] == a:
            count_min += 1
    counter[MIN] = max(counter[MIN], count_min)
    counter[MAX] = min(counter[MAX], count_max)
    if counter[MIN] > counter[MAX]:
        return PROP_INCONSISTENCY
    if count_min == count_max:
        return PROP_ENTAILMENT
    if count_min == counter[MAX]:  # we cannot have more domains equal to a
        for x_i in x:
            if x_i[MIN] == a and x_i[MAX] > a:
                x_i[MIN] = a + 1
            elif x_i[MIN] < a and x_i[MAX] == a:
                x_i[MAX] = a - 1
    if count_max == counter[MIN]:  # we cannot have more domains different from a
        for x_i in x:
            if x_i[MIN] <= a <= x_i[MAX]:
                x_i[:] = a
    return PROP_CONSISTENCY



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
import numpy as np
from numba import njit  # type: ignore
from numpy.typing import NDArray

from nucs.constants import EVENT_MASK_MIN_MAX, MAX, MIN, PROP_CONSISTENCY, PROP_ENTAILMENT, PROP_INCONSISTENCY


def get_complexity_equiv_eq(n: int, parameters: NDArray) -> float:
    """
    Returns the time complexity of the propagator as a float.
    :param n: the number of variables
    :param parameters: the parameters, unused here
    :return: a float
    """
    return 1


def get_triggers_equiv_eq(n: int, parameters: NDArray) -> NDArray:
    """
    Returns the triggers for this propagator.
    :param n: the number of variables
    :param parameters: the parameters, unused here
    :return: an array of triggers
    """
    return np.full(n, dtype=np.uint8, fill_value=EVENT_MASK_MIN_MAX)


@njit(cache=True)
def compute_domains_equiv_eq(domains: NDArray, parameters: NDArray) -> int:
    """
    Implements b <=> x = c.
    :param domains: the domains of the variables, b is the first domain, x is the second domain
    :param parameters: c is the first parameter
    :return: the status of the propagation (consistency, inconsistency or entailment) as an int
    """
    b = domains[0]
    x = domains[1]
    c = parameters[0]
    if b[MIN] == 0 and b[MAX] == 0:
        if x[MIN] == c:
            x[MIN] = c + 1
            if x[MIN] > x[MAX]:
                return PROP_INCONSISTENCY
        if x[MAX] == c:
            x[MAX] = c - 1
            if x[MIN] > x[MAX]:
                return PROP_INCONSISTENCY
    elif b[MIN] == 1 and b[MAX] == 1:
        if c < x[MIN] or c > x[MAX]:
            return PROP_INCONSISTENCY
        else:
            x[:] = c
            return PROP_ENTAILMENT
    if x[MIN] > c or x[MAX] < c:
        b[:] = 0
        return PROP_ENTAILMENT
    elif x[MIN] == c and x[MAX] == c:
        b[:] = 1
        return PROP_ENTAILMENT
    return PROP_CONSISTENCY



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
import logging
from abc import abstractmethod
from multiprocessing import Queue

logger = logging.getLogger(__name__)


class QueueSolver:
    """
    A solver which uses a queue.
    """

    @abstractmethod
    def minimize_and_queue(self, variable_idx: int, processor_idx: int, solution_queue: Queue, mode: str) -> None:
        """
        Enqueues the solution that minimizes a variable.
        :param variable_idx: the index of the variable to minimize
        :param processor_idx: the index of the processor running the minimizer
        :param solution_queue: the solution queue
        """
        ...

    @abstractmethod
    def maximize_and_queue(self, variable_idx: int, processor_idx: int, solution_queue: Queue, mode: str) -> None:
        """
        Enqueues the solution that maximizes a variable.
        :param variable_idx: the index of the variable to maximizer
        :param processor_idx: the index of the processor running the maximizer
        :param solution_queue: the solution queue
        """
        ...

    @abstractmethod
    def solve_and_queue(self, processor_idx: int, solution_queue: Queue) -> None:
        """
        Enqueues the solutions.
        :param processor_idx: the index of the processor
        :param solution_queue: the solution queue
        """
        ...



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
import numpy as np

from nucs.constants import PROP_CONSISTENCY, PROP_ENTAILMENT, PROP_INCONSISTENCY
from nucs.numpy_helper import new_parameters_by_values, new_shr_domains_by_values
from nucs.propagators.relation_propagator import compute_domains_relation


class TestRelation:
    def test_compute_domains_1(self) -> None:
        domains = new_shr_domains_by_values([(-5, 5), (-5, 5)])
        data = new_parameters_by_values([0, 7, 1, 4, 2, -7, 3, 3])
        assert compute_domains_relation(domains, data) == PROP_CONSISTENCY
        assert np.all(domains == np.array([[1, 3], [3, 4]]))

    def test_compute_domains_2(self) -> None:
        domains = new_shr_domains_by_values([(0, 3), (0, 3), (1, 8)])
        data = new_parameters_by_values(
            [0, 1, 0, 0, 2, 0, 0, 3, 0, 1, 1, 1, 1, 2, 2, 1, 3, 3, 2, 1, 2, 2, 2, 4, 2, 3, 6, 3, 1, 3, 3, 2, 6, 3, 3, 9]
        )
        assert compute_domains_relation(domains, data) == PROP_CONSISTENCY
        assert np.all(domains == np.array([[1, 3], [1, 3], [1, 6]]))

    def test_compute_domains_3(self) -> None:
        domains = new_shr_domains_by_values([(0, 3), (0, 3)])
        data = new_parameters_by_values([4, 5])
        assert compute_domains_relation(domains, data) == PROP_INCONSISTENCY

    def test_compute_domains_4(self) -> None:
        domains = new_shr_domains_by_values([(0, 3), (0, 3)])
        data = new_parameters_by_values([1, 2])
        assert compute_domains_relation(domains, data) == PROP_ENTAILMENT
        assert np.all(domains == np.array([[1, 1], [2, 2]]))

    def test_compute_domains_5(self) -> None:
        data = new_parameters_by_values([0, 1, 0, 0, 2, 1, 0, 3, 2, 1, 2, 3, 1, 3, 4, 2, 3, 5])
        assert compute_domains_relation(new_shr_domains_by_values([0, 1, (0, 5)]), data) == PROP_ENTAILMENT
        assert compute_domains_relation(new_shr_domains_by_values([0, 2, (0, 5)]), data) == PROP_ENTAILMENT
        assert compute_domains_relation(new_shr_domains_by_values([0, 3, (0, 5)]), data) == PROP_ENTAILMENT
        assert compute_domains_relation(new_shr_domains_by_values([2, 3, (0, 5)]), data) == PROP_ENTAILMENT
        assert compute_domains_relation(new_shr_domains_by_values([1, 3, (0, 5)]), data) == PROP_ENTAILMENT
        assert compute_domains_relation(new_shr_domains_by_values([1, 2, (0, 5)]), data) == PROP_ENTAILMENT



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
import numpy as np

from nucs.constants import PROP_CONSISTENCY, PROP_ENTAILMENT, PROP_INCONSISTENCY
from nucs.numpy_helper import new_parameters_by_values, new_shr_domains_by_values
from nucs.propagators.count_eq_propagator import compute_domains_count_eq


class TestCountEQ:
    def test_compute_domains_1(self) -> None:
        domains = new_shr_domains_by_values([(1, 4), (3, 5), (3, 6), (6, 8), 3, 5, 1])
        data = new_parameters_by_values([5])
        assert compute_domains_count_eq(domains, data) == PROP_CONSISTENCY
        assert np.all(domains == np.array([[1, 4], [3, 4], [3, 6], [6, 8], [3, 3], [5, 5], [1, 1]]))

    def test_compute_domains_2(self) -> None:
        domains = new_shr_domains_by_values([(1, 4), (3, 5), (3, 6), (6, 8), 3, 5, 2])
        data = new_parameters_by_values([5])
        assert compute_domains_count_eq(domains, data) == PROP_CONSISTENCY
        assert np.all(domains == np.array([[1, 4], [3, 5], [3, 6], [6, 8], [3, 3], [5, 5], [2, 2]]))

    def test_compute_domains_3(self) -> None:
        domains = new_shr_domains_by_values([(1, 4), (3, 5), (3, 6), (6, 8), 3, 5, 0])
        data = new_parameters_by_values([5])
        assert compute_domains_count_eq(domains, data) == PROP_INCONSISTENCY

    def test_compute_domains_4(self) -> None:
        domains = new_shr_domains_by_values([(1, 4), 5, (3, 6), (6, 8), 3, 5, (1, 2)])
        data = new_parameters_by_values([5])
        assert compute_domains_count_eq(domains, data) == PROP_CONSISTENCY
        assert np.all(domains == np.array([[1, 4], [5, 5], [3, 6], [6, 8], [3, 3], [5, 5], [2, 2]]))

    def test_compute_domains_5(self) -> None:
        domains = new_shr_domains_by_values([(1, 4), (3, 5), (3, 6), (6, 8), 3, 5, (-1, 10)])
        data = new_parameters_by_values([5])
        assert compute_domains_count_eq(domains, data) == PROP_CONSISTENCY
        assert np.all(domains == np.array([[1, 4], [3, 5], [3, 6], [6, 8], [3, 3], [5, 5], [1, 3]]))

    def test_compute_domains_6(self) -> None:
        domains = new_shr_domains_by_values([2, (0, 1), (3, 4), 2, 2, (2, 4)])
        data = new_parameters_by_values([2])
        assert compute_domains_count_eq(domains, data) == PROP_ENTAILMENT
        assert np.all(domains == np.array([[2, 2], [0, 1], [3, 4], [2, 2], [2, 2], [3, 3]]))



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
import pytest

from nucs.constants import STATS_IDX_SOLVER_SOLUTION_NB
from nucs.examples.magic_sequence.magic_sequence_problem import MagicSequenceProblem
from nucs.heuristics.heuristics import DOM_HEURISTIC_MIN_VALUE, VAR_HEURISTIC_FIRST_NOT_INSTANTIATED
from nucs.solvers.backtrack_solver import BacktrackSolver


class TestMagicSequence:

    @pytest.mark.parametrize("size,zero_nb", [(50, 46), (100, 96), (200, 196)])
    def test_magic_sequence(self, size: int, zero_nb: int) -> None:
        problem = MagicSequenceProblem(size)
        solver = BacktrackSolver(
            problem,
            decision_domains=list(range(size - 1, -1, -1)),
            var_heuristic_idx=VAR_HEURISTIC_FIRST_NOT_INSTANTIATED,
            dom_heuristic_idx=DOM_HEURISTIC_MIN_VALUE,
            stacks_max_height=128,
        )
        solutions = solver.find_all()
        assert solver.statistics[STATS_IDX_SOLVER_SOLUTION_NB] == 1
        assert solutions[0][0] == zero_nb



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
import math

import numpy as np
from numba import njit  # type: ignore
from numpy.typing import NDArray

from nucs.constants import EVENT_MASK_MIN_MAX, MAX, MIN, PROP_CONSISTENCY, PROP_INCONSISTENCY
from nucs.propagators.alldifferent_propagator import path_max, path_min, path_set


def get_complexity_gcc(n: int, parameters: NDArray) -> float:
    """
    Returns the time complexity of the propagator as a float.
    :param n: the number of variables
    :param parameters: the parameters, unused here
    :return: a float
    """
    return 2 * n * math.log(n) + 15 * n


def get_triggers_gcc(n: int, parameters: NDArray) -> NDArray:
    """
    This propagator is triggered whenever there is a change in the domain of a variable.
    :param n: the number of variables
    :return: an array of triggers
    """
    return np.full(n, dtype=np.uint8, fill_value=EVENT_MASK_MIN_MAX)


@njit(cache=True)
def init_partial_sum(first_value: int, m: int, values: NDArray) -> NDArray:
    """
    Inits the partial_sum data structure:
    ---------------------
    | sm | first_value |
    ---------------------
    | ds  | last_value  |
    ---------------------
    """
    partial_sum = np.zeros((2, m + 6), dtype=np.int32)
    partial_sum[0, -1] = first_value - 3
    partial_sum[1, -1] = first_value + m + 1
    sm = partial_sum[0, :-1]
    sm[0] = 0
    sm[1] = 1
    sm[2] = 2
    for i in range(2, m + 2):
        sm[i + 1] = sm[i] + values[i - 2]
    sm[m + 3] = sm[m + 2] + 1
    sm[m + 4] = sm[m + 3] + 1
    ds = partial_sum[1, :-1]
    i = m + 3
    j = m + 4
    while i > 0:
        while sm[i] == sm[i - 1]:
            ds[i] = j
            i -= 1
        ds[j] = i
        j = i
        i -= 1
    ds[j] = 0
    return partial_sum


@njit(cache=True)
def get_sum(psum: NDArray, start: int, end: int) -> int:
    fv = psum[0, -1]
    sum = psum[0, :-1]
    if start <= end:
        # assert fv <= start
        # assert end <= get_last_value(psum)
        return sum[end - fv] - sum[start - fv - 1]
    else:
        # assert fv <= end
        # assert start <= get_last_value(psum)
        return sum[end - fv - 1] - sum[start - fv]


@njit(cache=True)
def get_min_value(psum: NDArray) -> int:
    return psum[0, -1] + 3


@njit(cache=True)
def get_max_value(psum: NDArray) -> int:
    return psum[1, -1] - 2


@njit(cache=True)
def skip_non_null_elements_right(psum: NDArray, value: int) -> int:
    value -= psum[0, -1]
    return (value if psum[1, value] < value else psum[1, value]) + psum[0, -1]


@njit(cache=True)
def skip_non_null_elements_left(psum: NDArray, value: int) -> int:
    value -= psum[0, -1]
    return (psum[1, psum[1, value]] if psum[1, value] > value else value) + psum[0, -1]


@njit(cache=True)
def update_bounds(
    bounds: NDArray,
    n: int,
    domains: NDArray,
    ranks: NDArray,
    min_sorted_vars: NDArray,
    max_sorted_vars: NDArray,
    l: NDArray,
    u: NDArray,
) -> int:
    min_value = domains[min_sorted_vars[0], MIN]
    max_value = domains[max_sorted_vars[0], MAX] + 1
    last = l[0, -1] + 1
    bounds[0] = last
    i = j = nb = 0
    while True:
        if i < n and min_value <= max_value:
            if min_value != last:
                nb += 1
                bounds[nb] = last = min_value
            ranks[min_sorted_vars[i], MIN] = nb
            i += 1
            if i < n:
                min_value = domains[min_sorted_vars[i], MIN]
        else:
            if max_value != last:
                nb += 1
                bounds[nb] = last = max_value
            ranks[max_sorted_vars[j], MAX] = nb
            j += 1
            if j == n:
                break
            max_value = domains[max_sorted_vars[j], MAX] + 1
    bounds[nb + 1] = u[1, -1] + 1
    return nb


@njit(cache=True)
def filter_lower_max(
    n: int,
    nb: int,
    t: NDArray,
    d: NDArray,
    h: NDArray,
    bounds: NDArray,
    domains: NDArray,
    ranks: NDArray,
    max_sorted_vars: NDArray,
    u: NDArray,
) -> bool:
    for i in range(1, nb + 2):
        t[i] = i - 1
        h[i] = i - 1
        d[i] = get_sum(u, bounds[i - 1], bounds[i] - 1)
    for i, max_sorted_vars_i in enumerate(max_sorted_vars):
        x = ranks[max_sorted_vars_i, MIN]
        y = ranks[max_sorted_vars_i, MAX]
        z = path_max(t, x + 1)
        j = t[z]
        d[z] -= 1
        if d[z] == 0:
            t[z] = z + 1
            z = path_max(t, t[z])
            t[z] = j
        if d[z] < get_sum(
            u, bounds[y], bounds[z] - 1
        ):  # moved above the path compression which is not the case in the paper
            return False
        path_set(t, x + 1, z, z)  # path compression
        if h[x] > x:
            w = path_max(h, h[x])
            domains[max_sorted_vars_i, MIN] = bounds[w]
            path_set(h, x, w, w)  # path compression
            # changes = 1
        if d[z] == get_sum(u, bounds[y], bounds[z] - 1):
            path_set(h, h[y], j - 1, y)  # mark hall interval
            h[y] = j - 1  # hall interval[bounds[j], bounds[y]]
    return True


@njit(cache=True)
def filter_upper_max(
    n: int,
    nb: int,
    t: NDArray,
    d: NDArray,
    h: NDArray,
    bounds: NDArray,
    domains: NDArray,
    ranks: NDArray,
    min_sorted_vars: NDArray,
    u: NDArray,
) -> bool:
    for i in range(0, nb + 1):
        t[i] = i + 1
        h[i] = i + 1
        d[i] = get_sum(u, bounds[i], bounds[i + 1] - 1)
    for i in range(n - 1, -1, -1):
        min_sorted_vars_i = min_sorted_vars[i]
        x = ranks[min_sorted_vars_i, MAX]
        y = ranks[min_sorted_vars_i, MIN]
        z = path_min(t, x - 1)
        j = t[z]
        d[z] -= 1
        if d[z] == 0:
            t[z] = z - 1
            z = path_min(t, t[z])
            t[z] = j
        if d[z] < get_sum(
            u, bounds[z], bounds[y] - 1
        ):  # moved above the path compression which is not the case in the paper
            return False
        path_set(t, x - 1, z, z)  # path compression
        if h[x] < x:
            w = path_min(h, h[x])
            domains[min_sorted_vars_i, MAX] = bounds[w] - 1
            path_set(h, x, w, w)  # path compression
            # changes = 1
        if d[z] == get_sum(u, bounds[z], bounds[y] - 1):
            path_set(h, h[y], j + 1, y)  # mark hall interval
            h[y] = j + 1  # hall interval[bounds[j], bounds[y]]
    return True


@njit(cache=True)
def filter_lower_min(
    n: int,
    nb: int,
    tl: NDArray,
    c: NDArray,
    sets: NDArray,
    bounds: NDArray,
    domains: NDArray,
    ranks: NDArray,
    max_sorted_vars: NDArray,
    l: NDArray,
    stbl_intervals: NDArray,
    pot_stbl_sets: NDArray,
    new_mins: NDArray,
) -> bool:
    w = nb + 1
    for i in range(nb + 1, 0, -1):
        pot_stbl_sets[i] = i - 1
        stbl_intervals[i] = i - 1
        c[i] = get_sum(l, bounds[i - 1], bounds[i] - 1)
        if c[i] == 0:  # if the capacity between both bounds is zero, we have an unstable set between these two bounds
            sets[i - 1] = w
        else:
            sets[w] = i - 1
            w = i - 1
    w = nb + 1
    for i in range(nb + 1, -1, -1):
        if c[i] == 0:
            tl[i] = w
        else:
            tl[w] = i
            w = i
    for i, max_sorted_vars_i in enumerate(max_sorted_vars):  # visit intervals in increasing max order
        x = ranks[max_sorted_vars_i, MIN]
        y = ranks[max_sorted_vars_i, MAX]
        z = path_max(tl, x + 1)
        j = tl[z]
        if z != x + 1:
            # If bounds[z] - 1 belongs to a stable set, [bounds[x], bounds[z]) is a sub set of this stable set.
            w = path_max(pot_stbl_sets, x + 1)
            v = pot_stbl_sets[w]
            path_set(pot_stbl_sets, x + 1, w, w)  # path compression
            w = min(y, z)
            path_set(pot_stbl_sets, pot_stbl_sets[w], v, w)
            pot_stbl_sets[w] = v
        if c[z] <= get_sum(l, bounds[y], bounds[z] - 1):
            # (potentialStableSets[y], y] is a stable set
            w = path_max(stbl_intervals, pot_stbl_sets[y])
            path_set(stbl_intervals, pot_stbl_sets[y], w, w)  # path compression
            v = stbl_intervals[w]
            path_set(stbl_intervals, stbl_intervals[y], v, y)
            stbl_intervals[y] = v
        else:
            c[z] -= 1  # decrease the capacity between the two bounds
            if c[z] == 0:
                tl[z] = z + 1
                z = path_max(tl, tl[z])
                tl[z] = j
            # If the lower bound belongs to an unstable or a stable set, remind the new value we might assign to
            # the lower bound in case the variable doesn't belong to a stable set.
            if sets[x] > x:
                w = path_max(sets, x)
                new_mins[i] = w
                path_set(sets, x, w, w)  # path compression
            else:
                new_mins[i] = x  # do not shrink the variable
            if c[z] == get_sum(l, bounds[y], bounds[z] - 1):  # if an unstable set is discovered
                if sets[y] > y:  # consider stable and unstable sets beyond y
                    y = sets[y]  # equivalent to pathmax since the path is fully compressed
                path_set(sets, sets[y], j - 1, y)  # mark the new unstable set
                sets[y] = j - 1
        path_set(tl, x + 1, z, z)  # path compression
    if sets[nb] != 0:  # if there is a failure set
        return False
    # Perform path compression over all elements in the stable interval data structure. This data structure will no
    # longer be modified and will be accessed n or 2n times. Therefore, we can afford a linear time compression.
    for i in range(nb + 1, 0, -1):
        if stbl_intervals[i] > i:
            stbl_intervals[i] = w
        else:
            w = i
    # For all variables that are not a subset of a stable set, shrink the lower bound.
    for i in range(n - 1, -1, -1):
        max_sorted_vars_i = max_sorted_vars[i]
        x = ranks[max_sorted_vars_i, MIN]
        y = ranks[max_sorted_vars_i, MAX]
        if stbl_intervals[x] <= x or y > stbl_intervals[x]:
            domains[max_sorted_vars_i, MIN] = skip_non_null_elements_right(l, bounds[new_mins[i]])
            # changes = 1
    return True


@njit(cache=True)
def filter_upper_min(
    n: int,
    nb: int,
    tl: NDArray,
    c: NDArray,
    sets: NDArray,
    bounds: NDArray,
    domains: NDArray,
    ranks: NDArray,
    min_sorted_vars: NDArray,
    l: NDArray,
    stbl_intervals: NDArray,
    new_maxs: NDArray,
) -> bool:
    w = 0
    for i in range(nb + 1):
        c[i] = get_sum(l, bounds[i], bounds[i + 1] - 1)
        if c[i] == 0:  # if the capacity between both bounds is zero, we have an unstable set between these two bounds
            tl[i] = w
        else:
            tl[w] = i
            w = i
    tl[w] = nb + 1
    w = 0
    for i in range(1, nb + 1):
        if c[i - 1] == 0:
            sets[i] = w
        else:
            sets[w] = i
            w = i
    sets[w] = nb + 1
    for i in range(n - 1, -1, -1):  # visit intervals in decreasing max order
        min_sorted_vars_i = min_sorted_vars[i]
        x = ranks[min_sorted_vars_i, MAX]
        y = ranks[min_sorted_vars_i, MIN]
        # solve the lower bound problem
        z = path_min(tl, x - 1)
        j = tl[z]
        # If the variable is not in a discovered stable set
        # Possible optimization: use the array stbl_intervals to perform this test
        if c[z] > get_sum(l, bounds[z], bounds[y] - 1):
            c[z] -= 1
            if c[z] == 0:
                tl[z] = z - 1
                z = path_min(tl, tl[z])
                tl[z] = j
            if sets[x] < x:
                w = path_min(sets, sets[x])
                new_maxs[i] = w
                path_set(sets, x, w, w)  # path compression
            else:
                new_maxs[i] = x
            if c[z] == get_sum(l, bounds[z], bounds[y] - 1):
                if sets[y] < y:
                    y = sets[y]
                path_set(sets, sets[y], j + 1, y)  # loop
                sets[y] = j + 1
        path_set(tl, x - 1, z, z)
    #  For all variables that are not subsets of a stable set, shrink the lower bound.
    for i in range(n - 1, -1, -1):
        min_sorted_vars_i = min_sorted_vars[i]
        x = ranks[min_sorted_vars_i, MIN]
        y = ranks[min_sorted_vars_i, MAX]
        if stbl_intervals[x] <= x or y > stbl_intervals[x]:
            domains[min_sorted_vars_i, MAX] = skip_non_null_elements_left(l, bounds[new_maxs[i]] - 1)
            # changes = 1
    return True


@njit(cache=True)
def compute_domains_gcc(domains: NDArray, parameters: NDArray) -> int:
    """
    This propagator (Global Cardinality Constraint) enforces that l_j <= |{ i / x_i =v_j }| <= c_j for all j.
    Adapted from "An efficient bounds consistency algorithm for the global cardinality constraint".
    :param domains: the domains of the variables
    :param parameters: there are 1 + 2 * m parameters:
    the first domain value (v_0), then the m lower bounds, then the m upper bounds (capacities)
    """
    n = len(domains)
    m = (len(parameters) - 1) // 2  # number of values
    bounds_nb = 2 * n + 2
    ranks = np.zeros((n, 2), dtype=np.uint16)
    bounds = np.zeros(bounds_nb, dtype=np.int32)
    t = np.zeros(bounds_nb, dtype=np.uint16)  # critical capacity pointers
    d = np.zeros(bounds_nb, dtype=np.int32)  # differences between critical capacities
    h = np.zeros(bounds_nb, dtype=np.uint16)  # Hall interval pointers
    stbl_intervals = np.zeros(bounds_nb, dtype=np.int32)
    pot_stbl_sets = np.zeros(bounds_nb, dtype=np.int32)
    new_mins = np.zeros(n, dtype=np.int32)
    l = init_partial_sum(parameters[0], m, parameters[1 : 1 + m])
    u = init_partial_sum(parameters[0], m, parameters[1 + m :])
    min_sorted_vars = np.argsort(domains[:, MIN])
    max_sorted_vars = np.argsort(domains[:, MAX])
    nb = update_bounds(bounds, n, domains, ranks, min_sorted_vars, max_sorted_vars, l, u)
    # assert get_min_value(l) == get_min_value(u)
    # assert get_max_value(l) == get_max_value(u)
    # assert get_min_value(l) <= domains[min_sorted_vars[0], MIN]
    # assert domains[max_sorted_vars[n - 1], MAX] <= get_max_value(u)
    if get_sum(l, get_min_value(l), domains[min_sorted_vars[0], MIN] - 1) > 0:
        return PROP_INCONSISTENCY
    if get_sum(l, domains[max_sorted_vars[n - 1], MAX] + 1, get_max_value(l)) > 0:
        return PROP_INCONSISTENCY
    if not filter_lower_max(n, nb, t, d, h, bounds, domains, ranks, max_sorted_vars, u):
        return PROP_INCONSISTENCY
    if not filter_lower_min(
        n,
        nb,
        t,
        d,
        h,
        bounds,
        domains,
        ranks,
        max_sorted_vars,
        l,
        stbl_intervals,
        pot_stbl_sets,
        new_mins,
    ):
        return PROP_INCONSISTENCY
    if not filter_upper_max(n, nb, t, d, h, bounds, domains, ranks, min_sorted_vars, u):
        return PROP_INCONSISTENCY
    if not filter_upper_min(n, nb, t, d, h, bounds, domains, ranks, min_sorted_vars, l, stbl_intervals, new_mins):
        return PROP_INCONSISTENCY
    return PROP_CONSISTENCY



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
import logging
from abc import abstractmethod
from typing import Callable, Dict, Iterator, List, Optional

import numpy as np
from numba import njit  # type: ignore
from numpy.typing import NDArray

from nucs.constants import LOG_FORMAT, LOG_LEVEL_INFO, MAX, MIN
from nucs.problems.problem import Problem

logger = logging.getLogger(__name__)


class Solver:
    """
    A solver.
    """

    def __init__(self, problem: Optional[Problem], log_level: str = LOG_LEVEL_INFO):
        """
        Inits the solver.
        :param problem: a problem or None
        :param log_level: the log level as a string
        """
        logging.basicConfig(format=LOG_FORMAT, level=getattr(logging, log_level))  # TODO: move to examples
        logger.debug("Initializing Solver")
        if problem is not None:
            self.problem = problem
            problem.init()

    @abstractmethod
    def get_statistics(self) -> Dict[str, int]:
        """
        Returns the statistics as a dictionary.
        :return: a dictionary
        """
        ...

    @abstractmethod
    def solve(self) -> Iterator[NDArray]:
        """
        Returns an iterator over the solutions.
        :return: an iterator
        """
        ...

    def solve_all(self, func: Optional[Callable] = None) -> None:
        """
        Finds all solutions.
        :param func: a function to handle each solution found
        """
        for solution in self.solve():
            if func is not None:
                func(solution)

    def find_all(self) -> List[NDArray]:
        """
        Finds all solutions.
        :return: the list of all solutions
        """
        logger.info("Finding all solutions")
        solutions = []
        self.solve_all(lambda solution: solutions.append(solution))
        return solutions

    def find_one(self) -> Optional[NDArray]:
        logger.info("Finding one solution")
        for solution in self.solve():
            return solution
        return None

    @abstractmethod
    def minimize(self, variable_idx: int, mode: str) -> Optional[NDArray]:
        """
        Finds, if it exists, the solution to the problem that minimizes a given variable.
        :param variable_idx: the index of the variable
        :return: the solution if it exists or None
        """
        ...

    @abstractmethod
    def maximize(self, variable_idx: int, mode: str) -> Optional[NDArray]:
        """
        Finds, if it exists, the solution to the problem that maximizes a given variable.
        :param variable_idx: the index of the variable
        :return: the solution if it exists or None
        """
        ...


@njit(cache=True)
def get_solution(
    shr_domains_stack: NDArray, stacks_top: NDArray, dom_indices_arr: NDArray, dom_offsets_arr: NDArray
) -> NDArray:
    """
    Returns the solution to the problem.
    :param shr_domains_stack: the stack of shared domains
    :param stacks_top: the index of the top of the stacks as a Numpy array
    :param dom_indices_arr: the domain indices
    :param dom_offsets_arr: the domain offsets
    :return: a Numpy array
    """
    return shr_domains_stack[stacks_top[0], dom_indices_arr, MIN] + dom_offsets_arr


@njit(cache=True)
def is_solved(shr_domains_stack: NDArray, stacks_top: NDArray) -> bool:
    """
    Returns true iff the problem is solved.
    :param shr_domains_stack: the stack of shared domains
    :param stacks_top: the index of the top of the stacks as a Numpy array
    :return: a boolean
    """
    return bool(np.all(np.equal(shr_domains_stack[stacks_top[0], :, MIN], shr_domains_stack[stacks_top[0], :, MAX])))



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
from nucs.examples.knapsack.knapsack_problem import KnapsackProblem
from nucs.heuristics.heuristics import DOM_HEURISTIC_MAX_VALUE, VAR_HEURISTIC_FIRST_NOT_INSTANTIATED
from nucs.solvers.backtrack_solver import BacktrackSolver


class TestKnapsack:
    def test_knapsack(self) -> None:
        problem = KnapsackProblem(
            [40, 40, 38, 38, 36, 36, 34, 34, 32, 32, 30, 30, 28, 28, 26, 26, 24, 24, 22, 22],
            [40, 40, 38, 38, 36, 36, 34, 34, 32, 32, 30, 30, 28, 28, 26, 26, 24, 24, 22, 22],
            55,
        )
        solver = BacktrackSolver(
            problem, var_heuristic_idx=VAR_HEURISTIC_FIRST_NOT_INSTANTIATED, dom_heuristic_idx=DOM_HEURISTIC_MAX_VALUE
        )
        solution = solver.maximize(problem.weight)
        assert solution is not None
        assert solution[problem.weight] == 54



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
import pytest

from nucs.constants import STATS_IDX_SOLVER_SOLUTION_NB
from nucs.examples.quasigroup.quasigroup_problem import QuasigroupProblem
from nucs.heuristics.heuristics import DOM_HEURISTIC_SPLIT_LOW, VAR_HEURISTIC_SMALLEST_DOMAIN
from nucs.solvers.backtrack_solver import BacktrackSolver


class TestQuasigroup:
    @pytest.mark.parametrize(
        "kind, size, idempotent, solution_nb",
        [
            (3, 5, True, 0),
            (3, 6, True, 0),
            (3, 7, True, 0),
            (3, 8, True, 18),
            (4, 5, True, 3),
            (4, 6, True, 0),
            (4, 7, True, 0),
            (4, 8, True, 0),
            (5, 5, True, 1),
            (5, 6, True, 0),
            (5, 7, True, 3),
            (5, 8, True, 1),
            (5, 9, True, 0),
            (5, 10, True, 0),
            (5, 11, True, 5),
            # (5, 12, True, 0),      4s
            # (5, 13, True, 0),     97s
            # (5, 14, True, 0),   1380s
            # (5, 15, True, 0),   8h (6p)
            # (5, 16, True, 0),   ?
            # (5, 17, True, 0),   ?
            # (5, 18, True, 0),   ? open
            (6, 5, True, 0),
            (6, 6, True, 0),
            (6, 7, True, 0),
            (6, 8, True, 2),
            (6, 9, True, 4),
            (7, 5, True, 3),
            (7, 6, True, 0),
            (7, 7, True, 0),
            (7, 8, True, 0),
            # (7, 9, True, 1),
        ],
    )
    def test_qg(self, kind: int, size: int, idempotent: bool, solution_nb: int) -> None:
        problem = QuasigroupProblem(kind, size, idempotent, True)
        solver = BacktrackSolver(
            problem,
            decision_domains=list(range(0, size * size)),
            var_heuristic_idx=VAR_HEURISTIC_SMALLEST_DOMAIN,
            dom_heuristic_idx=DOM_HEURISTIC_SPLIT_LOW,
        )
        solver.solve_all()
        assert solver.statistics[STATS_IDX_SOLVER_SOLUTION_NB] == solution_nb



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
from nucs.problems.permutation_problem import PermutationProblem
from nucs.propagators.propagators import ALG_NO_SUB_CYCLE


class CircuitProblem(PermutationProblem):
    """
    A model for circuit.
    """

    def __init__(self, n: int):
        """
        Inits the circuit problem.
        :param n: the number of vertices
        """
        self.n = n
        super().__init__(n)
        self.shr_domains_lst[0] = [1, n - 1]
        self.shr_domains_lst[n - 1] = [0, n - 2]
        self.shr_domains_lst[n] = [1, n - 1]
        self.shr_domains_lst[2 * n - 1] = [0, n - 2]
        self.add_propagator((list(range(n)), ALG_NO_SUB_CYCLE, []))
        self.add_propagator((list(range(n, 2 * n)), ALG_NO_SUB_CYCLE, []))
        # self.add_propagator((list(range(n)), ALG_SCC, []))  # not worth the cost
        # self.add_propagator((list(range(n, 2*n)), ALG_SCC, []))  # not worth the cost



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
import numpy as np

from nucs.constants import PROP_CONSISTENCY
from nucs.numpy_helper import new_parameters_by_values, new_shr_domains_by_values
from nucs.propagators.gcc_propagator import compute_domains_gcc


class TestGCC:
    def test_compute_domains_0(self) -> None:
        domains = new_shr_domains_by_values([0])
        assert compute_domains_gcc(domains, new_parameters_by_values([0, 1, 1])) == PROP_CONSISTENCY
        assert np.all(domains == np.array([[0, 0]]))

    def test_compute_domains_1(self) -> None:
        domains = new_shr_domains_by_values([0, 1])
        assert compute_domains_gcc(domains, new_parameters_by_values([0, 1, 1, 1, 1])) == PROP_CONSISTENCY
        assert np.all(domains == np.array([[0, 0], [1, 1]]))

    def test_compute_domains_2(self) -> None:
        domains = new_shr_domains_by_values([0, (0, 1)])
        assert compute_domains_gcc(domains, new_parameters_by_values([0, 1, 1, 1, 1])) == PROP_CONSISTENCY
        assert np.all(domains == np.array([[0, 0], [1, 1]]))

    def test_compute_domains_3(self) -> None:
        domains = new_shr_domains_by_values([0, 2, (1, 2)])
        assert compute_domains_gcc(domains, new_parameters_by_values([0] + [1] * 6)) == PROP_CONSISTENCY
        assert np.all(domains == np.array([[0, 0], [2, 2], [1, 1]]))

    def test_compute_domains_4(self) -> None:
        domains = new_shr_domains_by_values([0, (0, 4), (0, 4), (0, 4), (0, 4)])
        assert compute_domains_gcc(domains, new_parameters_by_values([0] + [1] * 10)) == PROP_CONSISTENCY
        assert np.all(domains == np.array([[0, 0], [1, 4], [1, 4], [1, 4], [1, 4]]))

    def test_compute_domains_5(self) -> None:
        domains = new_shr_domains_by_values([(3, 6), (3, 4), (2, 5), (2, 4), (3, 4), (1, 6)])
        assert compute_domains_gcc(domains, new_parameters_by_values([1] + [1] * 12)) == PROP_CONSISTENCY
        assert np.all(domains == np.array([[6, 6], [3, 4], [5, 5], [2, 2], [3, 4], [1, 1]]))

    def test_compute_domains_6(self) -> None:
        domains = new_shr_domains_by_values([(3, 4), (2, 4), (3, 4), (2, 5), (3, 6), (1, 6)])
        assert compute_domains_gcc(domains, new_parameters_by_values([1] + [0] * 6 + [1] * 6)) == PROP_CONSISTENCY
        assert np.all(domains == np.array([[3, 4], [2, 2], [3, 4], [5, 5], [6, 6], [1, 1]]))



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
import numpy as np

from nucs.constants import PROP_CONSISTENCY, PROP_ENTAILMENT, PROP_INCONSISTENCY
from nucs.numpy_helper import new_parameters_by_values, new_shr_domains_by_values
from nucs.propagators.max_leq_propagator import compute_domains_max_leq


class TestMaxLEQ:
    def test_compute_domains_1(self) -> None:
        domains = new_shr_domains_by_values([(1, 4), (2, 5), (0, 2)])
        data = new_parameters_by_values([])
        assert compute_domains_max_leq(domains, data) == PROP_CONSISTENCY
        assert np.all(domains == np.array([[1, 2], [2, 2], [2, 2]]))

    def test_compute_domains_2(self) -> None:
        domains = new_shr_domains_by_values([(1, 4), (3, 5), (0, 2)])
        data = new_parameters_by_values([])
        assert compute_domains_max_leq(domains, data) == PROP_INCONSISTENCY

    def test_compute_domains_3(self) -> None:
        domains = new_shr_domains_by_values([(2, 4), (2, 5), (0, 1)])
        data = new_parameters_by_values([])
        assert compute_domains_max_leq(domains, data) == PROP_INCONSISTENCY

    def test_compute_domains_4(self) -> None:
        domains = new_shr_domains_by_values([(0, 1), (0, 1), (2, 3)])
        data = new_parameters_by_values([])
        assert compute_domains_max_leq(domains, data) == PROP_ENTAILMENT
        assert np.all(domains == np.array([[0, 1], [0, 1], [2, 3]]))

    def test_compute_domains_5(self) -> None:
        domains = new_shr_domains_by_values([(0, 1), (0, 1), (0, 0)])
        data = new_parameters_by_values([])
        assert compute_domains_max_leq(domains, data) == PROP_CONSISTENCY
        assert np.all(domains == np.array([[0, 0], [0, 0], [0, 0]]))



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
from nucs.heuristics.heuristics import DOM_HEURISTIC_MIN_VALUE
from nucs.problems.problem import Problem
from nucs.solvers.backtrack_solver import BacktrackSolver


class TestMinValueDomHeuristic:
    def test_find_all(self) -> None:
        problem = Problem([(1, 5)])
        solver = BacktrackSolver(problem, dom_heuristic_idx=DOM_HEURISTIC_MIN_VALUE)
        solutions = solver.find_all()
        assert len(solutions) == 5
        assert solutions == [[1], [2], [3], [4], [5]]



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
import numpy as np
from numba import njit  # type: ignore
from numpy.typing import NDArray

from nucs.constants import (
    EVENT_MASK_MAX,
    EVENT_MASK_MIN,
    MAX,
    MIN,
    PROP_CONSISTENCY,
    PROP_ENTAILMENT,
    PROP_INCONSISTENCY,
)


def get_complexity_affine_leq(n: int, parameters: NDArray) -> float:
    """
    Returns the time complexity of the propagator as a float.
    :param n: the number of variables
    :param parameters: the parameters, unused here
    :return: a float
    """
    return 2 * n


def get_triggers_affine_leq(n: int, parameters: NDArray) -> NDArray:
    """
    Returns the triggers for this propagator.
    :param n: the number of variables
    :param parameters: the parameters
    :return: an array of triggers
    """
    triggers = np.zeros(n, dtype=np.uint8)
    for i, c in enumerate(parameters[:-1]):
        if c < 0:
            triggers[i] = EVENT_MASK_MAX
        elif c > 0:
            triggers[i] = EVENT_MASK_MIN
    return triggers


@njit(cache=True)
def compute_domains_affine_leq(domains: NDArray, parameters: NDArray) -> int:
    """
    Implements Sigma_i a_i * x_i <= a_{n-1}.
    :param domains: the domains of the variables, x is an alias for domains
    :param parameters: the parameters of the propagator, a is an alias for parameters
    :return: the status of the propagation (consistency, inconsistency or entailment) as an int
    """
    domain_sum_min = domain_sum_max = parameters[-1]
    for i, c in enumerate(parameters[:-1]):
        if c > 0:
            domain_sum_min -= c * domains[i, MAX]
            domain_sum_max -= c * domains[i, MIN]
        elif c < 0:
            domain_sum_min -= c * domains[i, MIN]
            domain_sum_max -= c * domains[i, MAX]
    if domain_sum_min >= 0:
        return PROP_ENTAILMENT
    old_domains = np.copy(domains)
    for i, c in enumerate(parameters[:-1]):
        if c != 0:
            if c > 0:
                new_max = old_domains[i, MIN] + (domain_sum_max // c)
                domains[i, MAX] = min(domains[i, MAX], new_max)
            else:
                new_min = old_domains[i, MAX] - (-domain_sum_max // c)
                domains[i, MIN] = max(domains[i, MIN], new_min)
            if domains[i, MIN] > domains[i, MAX]:
                return PROP_INCONSISTENCY
    return PROP_CONSISTENCY



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
import numpy as np
from numba import njit  # type: ignore
from numpy.typing import NDArray

from nucs.constants import EVENT_MASK_MIN_MAX, MAX, MIN, PROP_CONSISTENCY, PROP_INCONSISTENCY


def get_complexity_scc(n: int, parameters: NDArray) -> float:
    """
    Returns the time complexity of the propagator as a float.
    :param n: the number of variables
    :param parameters: the parameters, unused here
    :return: a float
    """
    return 2 * n * n


def get_triggers_scc(n: int, parameters: NDArray) -> NDArray:
    """
    Returns the triggers for this propagator.
    :param n: the number of variables
    :param parameters: the parameters, unused here
    :return: an array of triggers
    """
    return np.full(n, dtype=np.uint8, fill_value=EVENT_MASK_MIN_MAX)


@njit(cache=False)  # Numba issue with cached recursive functions
def dfs_row(n: int, graph: NDArray, i: int, visited: NDArray) -> None:
    visited[i] = True
    for j in range(n):
        if graph[i, j] and not visited[j]:
            dfs_row(n, graph, j, visited)


@njit(cache=False)  # Numba issue with cached recursive functions
def dfs_col(n: int, graph: NDArray, j: int, visited: NDArray) -> None:
    visited[j] = True
    for i in range(n):
        if graph[i, j] and not visited[i]:
            dfs_col(n, graph, i, visited)


@njit(cache=False)  # Numba issue with recursive functions
def compute_domains_scc(domains: NDArray, parameters: NDArray) -> int:
    """
    :param domains: the domains of the variables
    :param parameters: unused here
    :return: the status of the propagation (consistency, inconsistency or entailment) as an int
    """
    n = len(domains)
    graph = np.zeros((n, n), dtype=np.bool)
    for i in range(n):
        graph[i, domains[i, MIN] : (domains[i, MAX] + 1)] = True
    visited = np.zeros(n, dtype=np.bool)
    dfs_row(n, graph, 0, visited)
    if not np.all(visited):
        return PROP_INCONSISTENCY
    visited[:] = False
    dfs_col(n, graph, 0, visited)
    if not np.all(visited):
        return PROP_INCONSISTENCY
    return PROP_CONSISTENCY



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
import pytest

from nucs.constants import STATS_IDX_SOLVER_SOLUTION_NB
from nucs.problems.permutation_problem import PermutationProblem
from nucs.solvers.backtrack_solver import BacktrackSolver


class TestPermutationProblem:
    @pytest.mark.parametrize("size, solution_nb", [(2, 2), (3, 6), (4, 24)])
    def test_solve_all(self, size: int, solution_nb: int) -> None:
        problem = PermutationProblem(size)
        solver = BacktrackSolver(problem)
        solver.solve_all()
        assert solver.statistics[STATS_IDX_SOLVER_SOLUTION_NB] == solution_nb

    def test_solve(self) -> None:
        problem = PermutationProblem(5)
        solver = BacktrackSolver(problem)
        it = solver.solve()
        solution = next(it)
        assert solution.tolist() == [0, 1, 2, 3, 4, 0, 1, 2, 3, 4]
        solution = next(it)
        assert solution.tolist() == [0, 1, 2, 4, 3, 0, 1, 2, 4, 3]



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
from typing import Callable

from nucs.solvers.bound_consistency_algorithm import bound_consistency_algorithm
from nucs.solvers.shaving_consistency_algorithm import shaving_consistency_algorithm

CONSISTENCY_ALG_FCTS = []


def register_consistency_algorithm(consistency_algorithm_fct: Callable) -> int:
    """
    Register a consistency algorithm by adding its function to the corresponding list of functions.
    :param consistency_algorithm_fct: a function that enforces consistency
    :return: the index of the consistency algorithm
    """
    CONSISTENCY_ALG_FCTS.append(consistency_algorithm_fct)
    return len(CONSISTENCY_ALG_FCTS) - 1


CONSISTENCY_ALG_BC = register_consistency_algorithm(bound_consistency_algorithm)
CONSISTENCY_ALG_SHAVING = register_consistency_algorithm(shaving_consistency_algorithm)



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
import pytest

from nucs.constants import OPTIM_PRUNE, STATS_IDX_SOLVER_SOLUTION_NB
from nucs.examples.tsp.tsp_instances import TSP_INSTANCES
from nucs.examples.tsp.tsp_problem import TSPProblem
from nucs.examples.tsp.tsp_var_heuristic import tsp_var_heuristic
from nucs.heuristics.heuristics import DOM_HEURISTIC_MIN_COST, register_var_heuristic
from nucs.solvers.backtrack_solver import BacktrackSolver


class TestTSP:
    def test_tsp_1(self) -> None:
        problem = TSPProblem([[0, 2, 1, 2], [2, 0, 2, 1], [1, 2, 0, 2], [2, 1, 2, 0]])
        solver = BacktrackSolver(problem, decision_domains=[0, 1, 2, 3])
        solution = solver.minimize(problem.total_cost)
        assert solution is not None
        assert solution[:4].tolist() == [1, 3, 0, 2]
        assert solution[problem.total_cost] == 6
        assert solver.statistics[STATS_IDX_SOLVER_SOLUTION_NB] == 2

    @pytest.mark.parametrize(
        "name, minimum",
        [
            ("GR17", 2085),
            #  ("GR21", 2707),
            #  ("GR24", 1272)
        ],
    )
    def test_tsp_gr(self, name: str, minimum: int) -> None:
        tsp_instance = TSP_INSTANCES[name]
        n = len(tsp_instance)
        problem = TSPProblem(tsp_instance)
        tsp_var_heuristic_idx = register_var_heuristic(tsp_var_heuristic)
        costs = tsp_instance + tsp_instance
        solver = BacktrackSolver(
            problem,
            decision_domains=list(range(0, 2 * n)),
            var_heuristic_idx=tsp_var_heuristic_idx,
            var_heuristic_params=costs,
            dom_heuristic_idx=DOM_HEURISTIC_MIN_COST,
            dom_heuristic_params=costs,
        )
        solution = solver.minimize(problem.total_cost, mode=OPTIM_PRUNE)
        assert solution is not None
        assert solution[problem.total_cost] == minimum



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
from typing import Any

import numpy as np
import pytest

from nucs.constants import PROP_CONSISTENCY, PROP_ENTAILMENT, PROP_INCONSISTENCY, STATS_IDX_SOLVER_SOLUTION_NB
from nucs.numpy_helper import new_parameters_by_values, new_shr_domains_by_values
from nucs.problems.problem import Problem
from nucs.propagators.lexicographic_leq_propagator import compute_domains_lexicographic_leq
from nucs.propagators.propagators import ALG_LEXICOGRAPHIC_LEQ
from nucs.solvers.backtrack_solver import BacktrackSolver


class TestLexicographicLEQ:
    def test_compute_domains_1(self) -> None:
        domains = new_shr_domains_by_values([(0, 1), 0, 1, 1])
        data = new_parameters_by_values([])
        assert compute_domains_lexicographic_leq(domains, data) == PROP_ENTAILMENT
        assert np.all(domains == np.array([[0, 1], [0, 0], [1, 1], [1, 1]]))

    def test_compute_domains_2(self) -> None:
        domains = new_shr_domains_by_values([(0, 1), (0, 1), (0, 1), (0, 1)])
        data = new_parameters_by_values([])
        assert compute_domains_lexicographic_leq(domains, data) == PROP_CONSISTENCY
        assert np.all(domains == np.array([[0, 1], [0, 1], [0, 1], [0, 1]]))

    @pytest.mark.parametrize(
        "values,state",
        [
            ([0, 0, 0, 0], PROP_ENTAILMENT),
            ([0, 0, 0, 1], PROP_ENTAILMENT),
            ([0, 0, 1, 0], PROP_ENTAILMENT),
            ([0, 0, 1, 1], PROP_ENTAILMENT),
            ([0, 1, 0, 0], PROP_INCONSISTENCY),
            ([0, 1, 0, 1], PROP_ENTAILMENT),
            ([0, 1, 1, 0], PROP_ENTAILMENT),
            ([0, 1, 1, 1], PROP_ENTAILMENT),
            ([1, 0, 0, 0], PROP_INCONSISTENCY),
            ([1, 0, 0, 1], PROP_INCONSISTENCY),
            ([1, 0, 1, 0], PROP_ENTAILMENT),
            ([1, 0, 1, 1], PROP_ENTAILMENT),
            ([1, 1, 0, 0], PROP_INCONSISTENCY),
            ([1, 1, 0, 1], PROP_INCONSISTENCY),
            ([1, 1, 1, 0], PROP_INCONSISTENCY),
            ([1, 1, 1, 1], PROP_ENTAILMENT),
        ],
    )
    def test_compute_domains_values(self, values: Any, state: int) -> None:
        domains = new_shr_domains_by_values(values)
        data = new_parameters_by_values([])
        assert compute_domains_lexicographic_leq(domains, data) == state

    def test_solve_1(self) -> None:
        problem = Problem(
            shr_domains_lst=[(0, 1), (0, 1), (0, 1), (0, 1)],
            dom_indices_lst=[0, 1, 2, 3],
            dom_offsets_lst=[0, 0, 0, 0],
        )
        problem.add_propagator(([0, 1, 2, 3], ALG_LEXICOGRAPHIC_LEQ, []))
        solver = BacktrackSolver(problem)
        solver.solve_all()
        assert solver.statistics[STATS_IDX_SOLVER_SOLUTION_NB] == 10

    def test_solve_2(self) -> None:
        problem = Problem(
            shr_domains_lst=[(1, 1), (0, 1), (0, 1), (0, 1)],
            dom_indices_lst=[0, 1, 2, 3],
            dom_offsets_lst=[0, 0, 0, 0],
        )
        problem.add_propagator(([0, 1, 2, 3], ALG_LEXICOGRAPHIC_LEQ, []))
        solver = BacktrackSolver(problem)
        solver.solve_all()
        assert solver.statistics[STATS_IDX_SOLVER_SOLUTION_NB] == 3



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
from typing import List

from nucs.problems.circuit_problem import CircuitProblem
from nucs.propagators.propagators import ALG_AFFINE_EQ, ALG_ELEMENT_IV


class TSPProblem(CircuitProblem):
    """ """

    def __init__(self, costs: List[List[int]]) -> None:
        """
        Inits the problem.
        :param costs: the costs between vertices as a list of lists of integers
        """
        n = len(costs)
        super().__init__(n)
        max_costs = [max(cost_row) for cost_row in costs]
        min_costs = [min([cost for cost in cost_row if cost > 0]) for cost_row in costs]
        self.succ_costs = self.add_variables([(min_costs[i], max_costs[i]) for i in range(n)])
        self.pred_costs = self.add_variables([(min_costs[i], max_costs[i]) for i in range(n)])
        self.total_cost = self.add_variable((sum(min_costs), sum(max_costs)))  # the total cost
        self.add_propagators([([i, self.succ_costs + i], ALG_ELEMENT_IV, costs[i]) for i in range(n)])
        self.add_propagators([([n + i, self.pred_costs + i], ALG_ELEMENT_IV, costs[i]) for i in range(n)])
        self.add_propagator(
            (list(range(self.succ_costs, self.succ_costs + n)) + [self.total_cost], ALG_AFFINE_EQ, [1] * n + [-1, 0])
        )
        self.add_propagator(
            (list(range(self.pred_costs, self.pred_costs + n)) + [self.total_cost], ALG_AFFINE_EQ, [1] * n + [-1, 0])
        )
        # total_cost_prop_idx = register_propagator(
        #     get_triggers_total_cost, get_complexity_total_cost, compute_domains_total_cost
        # )
        # costs = [cost for cost_row in costs for cost in cost_row]
        # self.add_propagator((list(range(n)) + [self.total_cost], total_cost_prop_idx, costs))
        # self.add_propagator((list(range(n, 2 * n)) + [self.total_cost], total_cost_prop_idx, costs))



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
from nucs.constants import PROP_CONSISTENCY, PROP_INCONSISTENCY
from nucs.numpy_helper import new_parameters_by_values, new_shr_domains_by_values
from nucs.propagators.no_sub_cycle_propagator import compute_domains_no_sub_cycle


class TestNoSubCycle:
    def test_compute_domains_3_consistency(self) -> None:
        domains = new_shr_domains_by_values([(1, 1), (2, 2), (0, 0)])
        data = new_parameters_by_values([])
        assert compute_domains_no_sub_cycle(domains, data) == PROP_CONSISTENCY

    def test_compute_domains_3_inconsistency(self) -> None:
        domains = new_shr_domains_by_values([(1, 1), (0, 0), (2, 2)])
        data = new_parameters_by_values([])
        assert compute_domains_no_sub_cycle(domains, data) == PROP_INCONSISTENCY

    def test_compute_domains_single(self) -> None:
        domains = new_shr_domains_by_values([0])
        data = new_parameters_by_values([])
        assert compute_domains_no_sub_cycle(domains, data) == PROP_INCONSISTENCY



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
from nucs.problems.problem import Problem
from nucs.propagators.propagators import ALG_AFFINE_LEQ, ALG_EXACTLY_TRUE, ALG_LEXICOGRAPHIC_LEQ


class SchurLemmaProblem(Problem):
    """
    CSPLIB problem #15 - https://www.csplib.org/Problems/prob015/
    """

    def __init__(self, n: int, symmetry_breaking: bool = True) -> None:
        """
        Inits the problem.
        :param n: the size of the problem
        :param symmetry_breaking: a boolean indicating if symmetry constraints should be added to the model
        """
        super().__init__([(0, 1)] * n * 3)
        for x in range(n):
            self.add_propagator(([x * 3, x * 3 + 1, x * 3 + 2], ALG_EXACTLY_TRUE, [1]))
        for x in range(n):
            for y in range(n):
                z = (x + 1) + (y + 1) - 1
                if 0 <= z < n:
                    for k in range(3):
                        self.add_propagator(([3 * x + k, 3 * y + k, 3 * z + k], ALG_AFFINE_LEQ, [1, 1, 1, 2]))
        if symmetry_breaking:
            self.add_propagator(
                (
                    list(range(0, n * 3, 3)) + list(range(1, n * 3, 3)) + list(range(2, n * 3, 3)),
                    ALG_LEXICOGRAPHIC_LEQ,
                    [],
                )
            )
